@@ -1,8 +1,8 @@
 #------------------------------------------------------------------------------
 #$Author: antanas $
-#$Date: 2022-07-31 04:53:56 +0300 (Sun, 31 Jul 2022) $
-#$Revision: 9351 $
-#$URL: svn+ssh://www.crystallography.net/home/coder/svn-repositories/cod-tools/tags/v3.8.1/src/lib/perl5/COD/CIF/DDL.pm $
+#$Date: 2023-11-27 18:45:20 +0200 (Mon, 27 Nov 2023) $
+#$Revision: 9765 $
+#$URL: svn+ssh://www.crystallography.net/home/coder/svn-repositories/cod-tools/tags/v3.9.0/src/lib/perl5/COD/CIF/DDL.pm $
 #------------------------------------------------------------------------------
 #*
 #* A set of generic subroutines for DDL handling.
@@ -215,6 +215,7 @@ sub ddl1_to_ddlm
         _enumeration_range   => '_enumeration.range',
         _example             => '_description_example.case',
         _example_detail      => '_description_example.detail',
+        _list_reference      => '_category_key.name',
         _units_detail        => '_units.code',
     );
 
@@ -283,7 +284,7 @@ sub ddl1_to_ddlm
                          'in DDLm' . "\n";
                 }
 
-                # Uppercasing category names to make them stand out:
+                # Uppercase category names to make them stand out.
                 $name = uc $name;
                 set_tag( $ddl_datablock,
                          '_definition.class',
@@ -304,10 +305,14 @@ sub ddl1_to_ddlm
                     set_tag( $ddl_datablock,
                              '_type.contents',
                              $typemap{get_dic_item_value( $ddl_datablock, '_type' )} );
-                    if( get_dic_item_value( $ddl_datablock, '_type' ) eq 'numb' &&
-                        defined get_dic_item_value( $ddl_datablock, '_type_conditions' ) &&
-                        get_dic_item_value( $ddl_datablock, '_type_conditions' ) =~ /^esd|su$/ ) {
-                        $type_purpose = 'Measurand';
+                    if( get_dic_item_value( $ddl_datablock, '_type' ) eq 'numb' ) {
+                        if( defined get_dic_item_value( $ddl_datablock, '_type_conditions' ) &&
+                            get_dic_item_value( $ddl_datablock, '_type_conditions' ) =~ /^esd|su$/ ) {
+                            $type_purpose = 'Measurand';
+                        }
+                        if( !defined get_dic_item_value( $ddl_datablock, '_units_detail' ) ) {
+                            set_tag( $ddl_datablock, '_units.code', 'none' );
+                        }
                     }
                 }
 
@@ -316,9 +321,7 @@ sub ddl1_to_ddlm
                 }
 
                 if( $type_purpose ) {
-                    set_tag( $ddl_datablock,
-                             '_type.purpose',
-                             $type_purpose );
+                    set_tag( $ddl_datablock, '_type.purpose', $type_purpose );
                 }
             }
 
@@ -327,6 +330,14 @@ sub ddl1_to_ddlm
                 warn "'_units_detail' is not defined for '$ddl_datablock->{name}'\n";
             }
 
+            # Convert DDL1 parent item to DDLm linked item.
+            if ( defined $ddl_datablock->{'values'}{'_list_link_parent'} ) {
+                rename_tag( $ddl_datablock,
+                            '_list_link_parent',
+                            '_name.linked_item_id' );
+                set_tag( $ddl_datablock, '_type.purpose', 'Link' );
+            }
+
             for my $tag (sort keys %tags_to_rename) {
                 next if !defined get_dic_item_value( $ddl_datablock, $tag );
 
@@ -360,6 +371,14 @@ sub ddl1_to_ddlm
                          $category_overview );
             }
 
+            # Uppercase parent categories of other categories to stand out.
+            # Assume that _definition.scope is always set to 'Category'.
+            if (defined $ddl_datablock->{'values'}{'_definition.scope'}) {
+                set_tag( $ddl_datablock,
+                         '_name.category_id',
+                       uc $ddl_datablock->{'values'}{'_name.category_id'}[0] );
+            }
+
             set_tag( $ddl_datablock, '_definition.id', $name );
             set_tag( $ddl_datablock,
                      '_name.object_id',
@@ -369,42 +388,197 @@ sub ddl1_to_ddlm
         }
     }
 
+    my %data_name_to_frame =
+                        map { uc $_->{'values'}{'_definition.id'}[0] => $_ }
+                            @{$ddlm_datablock->{'save_blocks'}};
+
+    move_ddlm_keys_to_category_definitions($ddlm_datablock, \%data_name_to_frame);
+
+    for my $save_frame (@{$ddlm_datablock->{'save_blocks'}}) {
+        my $data_name = $save_frame->{'values'}{'_definition.id'}[0];
+        if (defined get_dic_item_value( $save_frame, '_list_mandatory' )) {
+            # TODO: consider is the value is set to 'yes' or 'no'.
+            # Normally, the value is only set to 'yes' as 'no' is the default.
+            warn "conversion of data item '$data_name' is lossy -- " .
+                 'constraints imposed by the \'_list_mandatory\' DDL1 ' .
+                 'attribute cannot be expressed in DDLm.' . "\n";
+            exclude_tag( $save_frame, '_list_mandatory' );
+        }
+
+        if ( defined get_dic_item_value( $save_frame, '_related_item' ) ||
+             defined get_dic_item_value( $save_frame, '_related_function' ) ) {
+            warn "conversion of data item '$data_name' is lossy -- " .
+                 'the \'_related_item\' and \'_related_function\' DDL1 ' .
+                 'attributes will not be automatically translated to DDLm.' .
+                 "\n";
+            for my $tag ( qw( _related_item _related_function ) ) {
+                next if !defined get_dic_item_value( $save_frame, $tag );
+                exclude_tag( $save_frame, $tag );
+            }
+        }
+
+        # This subroutine must be called after converting all of the
+        # '_list_link_parent' DDL1 attributes.
+        convert_ddl1_child_item_links( $save_frame, \%data_name_to_frame );
+    }
+
+    my $dic_version;
+    if (defined $new_version) {
+        $dic_version = $new_version
+    } else {
+        $dic_version = '0.0.1' . '+DDL1-version.' .
+                       $ddl_datablocks->[0]{'values'}{'_dictionary_version'}[0];
+    }
+
     set_tag( $ddlm_datablock, '_dictionary.title', $dictionary_name );
-    set_tag( $ddlm_datablock,
-             '_dictionary.version',
-             ($new_version
-                ? $new_version
-                : $ddl_datablocks->[0]{values}{_dictionary_version}[0]) );
+    set_tag( $ddlm_datablock, '_dictionary.version', $dic_version );
     set_tag( $ddlm_datablock, '_dictionary.date', $date );
     set_tag( $ddlm_datablock, '_dictionary.class', 'Instance' );
-    set_tag( $ddlm_datablock, '_dictionary.ddl_conformance', '3.13.1' );
+    set_tag( $ddlm_datablock, '_dictionary.ddl_conformance', '4.1.0' );
 
-    if( exists $ddl_datablocks->[0]{values}{_dictionary_history} ) {
+    if( exists $ddl_datablocks->[0]{'values'}{'_dictionary_history'} ) {
         set_loop_tag( $ddlm_datablock,
                       '_dictionary_audit.version',
                       '_dictionary_audit.version',
-                      [ $ddl_datablocks->[0]{values}{_dictionary_version}[0] ] );
+                      [ $dic_version ] );
         set_loop_tag( $ddlm_datablock,
                       '_dictionary_audit.date',
                       '_dictionary_audit.version',
-                      [ $ddl_datablocks->[0]{values}{_dictionary_update}[0] ] );
+                      [ $date ] );
+
+        my $old_log_message = $ddl_datablocks->[0]{'values'}
+                                                  {'_dictionary_history'}[0];
+        $old_log_message =~ s/^(?:[ \t]*\n)*|\s*$//gs;
+
+        my $new_log_message = "\n" .
+            'Automatically converted from DDL1 to DDLm.' .
+            "\n\n" .
+            'History of the original DDL1 dictionary:' .
+            "\n\n" .
+            $old_log_message;
+
         set_loop_tag( $ddlm_datablock,
                       '_dictionary_audit.revision',
                       '_dictionary_audit.version',
-                      $ddl_datablocks->[0]{values}{_dictionary_history} );
-        if( $new_version ) {
-            unshift @{$ddlm_datablock->{values}{'_dictionary_audit.version'}},
-                    $new_version;
-            unshift @{$ddlm_datablock->{values}{'_dictionary_audit.date'}},
-                    strftime( '%F', gmtime() );
-            unshift @{$ddlm_datablock->{values}{'_dictionary_audit.revision'}},
-                    'Automatically converting to DDLm';
-        }
+                      [ $new_log_message ] );
     }
 
     return $ddlm_datablock;
 }
 
+##
+# Moves the '_category_key.name' attribute from data item definitions to the
+# definitions of parent categories.
+#
+# Composite keys as well as key items from external dictionaries are currently
+# not supported and are silently left unmodified. Invalid definitions are also
+# silently left unmodified. 
+#
+# @param $ddlm_dic_block
+#       Reference to a DDLm dictionary data block as returned by the
+#       COD::CIF::Parser module.
+# @param $data_name_to_frame
+#       Reference to a hash data structure in which all definition save frames
+#       from a single dictionary are referenced by their main data name.
+##
+sub move_ddlm_keys_to_category_definitions
+{
+    my ($ddlm_dic_block, $data_name_to_frame) = @_;
+
+    # Group data items that share the same key.
+    my %key_item_to_loop_items;
+    for my $save_frame (@{$ddlm_dic_block->{'save_blocks'}}) {
+        next if !defined $save_frame->{'values'}{'_category_key.name'};
+        # TODO: implement the handling of composite keys.
+        next if @{$save_frame->{'values'}{'_category_key.name'}} > 1;
+        my $key_name = $save_frame->{'values'}{'_category_key.name'}[0];
+        push @{$key_item_to_loop_items{$key_name}}, $save_frame;
+    }
+
+    for my $key_item_name (sort keys %key_item_to_loop_items) {
+        my $key_item_block = $data_name_to_frame->{uc $key_item_name};
+
+        # TODO: handle key items that are defined in external dictionaries
+        # (e.g. _atom_site_label).
+        next if !defined $key_item_block;
+
+        my $category_name = $key_item_block->{'values'}{'_name.category_id'}[0];
+
+        # Safeguard against missing category definitions.
+        next if !defined $data_name_to_frame->{uc $category_name};
+        my $category_block = $data_name_to_frame->{uc $category_name};
+        # Avoid overriding existing category keys.
+        next if defined $category_block->{'values'}{'_category_key.name'};
+
+        my $move_key_to_category = 0;
+        for my $loop_item (@{$key_item_to_loop_items{$key_item_name}}) {
+            # Key item and loop item must share the same category. 
+            next if uc $loop_item->{'values'}{'_name.category_id'}[0] ne
+                    uc $category_name;
+            exclude_tag($loop_item, '_category_key.name');
+            $move_key_to_category = 1;
+        }
+        if ($move_key_to_category) {
+            set_tag( $category_block, '_category_key.name', $key_item_name );
+            set_tag( $category_block, '_definition.class', 'Loop' );
+            # A simple non-composite key item is mandatory in a loop
+            # and thus satisfies the '_list_mandatory' constraints.
+            next if !defined get_dic_item_value( $key_item_block,
+                                                 '_list_mandatory' );
+            exclude_tag( $key_item_block, '_list_mandatory' );
+        }
+    }
+
+    return;
+}
+
+##
+# Converts the '_list_link_child' DDL1 attribute to an equivalent DDLm
+# attribute. This subroutine assumes that the '_list_link_parent' DDL1
+# attribute in all save frames has already been converted to the DDLm
+# equivalent.
+#
+# @param $save_frame
+#       Reference to a save frame that contains a single item or category
+#       definition. The overall data structure is identical to that returned
+#       by the COD::CIF::Parser.
+# @param $data_name_to_frame
+#       Reference to a hash data structure in which all definition save frames
+#       from a single dictionary are referenced by their main data name.
+##
+sub convert_ddl1_child_item_links
+{
+    my ($save_frame, $data_name_to_frame) = @_;
+
+    return if !defined $save_frame->{'values'}{'_list_link_child'};
+    my @child_names = @{ $save_frame->{'values'}{'_list_link_child'} };
+    my $parent_name = $save_frame->{'values'}{'_definition.id'}[0];
+    for my $child_name (@child_names) {
+        if (defined $data_name_to_frame->{uc $child_name}) {
+            my $child_frame = $data_name_to_frame->{uc $child_name};
+            if (defined $child_frame->{'values'}{'_name.linked_item_id'}) {
+                my $current_parent_name =
+                            $child_frame->{'values'}{'_name.linked_item_id'}[0];
+                next if uc $current_parent_name eq uc $parent_name;
+                warn "unable to mark item '$child_name' as linked to item " .
+                     "'$parent_name' based on the value of the " .
+                     "\'_list_link_child\' DDL1 attribute -- item is already " .
+                     "linked to the '$current_parent_name' data item" . ".\n";
+            } else {
+                set_tag( $child_frame, '_name.linked_item_id', $parent_name )
+            }
+        } else {
+            warn "unable to mark item '$child_name' as linked to item " .
+                 "'$parent_name' based on the value of the " .
+                 "\'_list_link_child\' DDL1 attribute -- item could not be " .
+                 "located in the given dictionary" . ".\n";
+        }
+    }
+    exclude_tag( $save_frame, '_list_link_child' );
+
+    return;
+}
+
 ##
 # Converts (in a rather crude way) CIF data block to a DDLm dictionary.
 ##
