@@ -1,8 +1,8 @@
 #------------------------------------------------------------------------------
 #$Author: antanas $
-#$Date: 2022-09-19 10:31:20 +0300 (Mon, 19 Sep 2022) $
-#$Revision: 9405 $
-#$URL: svn+ssh://www.crystallography.net/home/coder/svn-repositories/cod-tools/tags/v3.8.0/src/lib/perl5/COD/CIF/Data/CIF2COD.pm $
+#$Date: 2024-03-04 11:46:13 +0200 (Mon, 04 Mar 2024) $
+#$Revision: 9897 $
+#$URL: svn+ssh://www.crystallography.net/home/coder/svn-repositories/cod-tools/tags/v3.9.0/src/lib/perl5/COD/CIF/Data/CIF2COD.pm $
 #------------------------------------------------------------------------------
 #*
 #  Parse a CIF file, prepare a COD database table entry from it.
@@ -36,8 +36,6 @@ our @EXPORT_OK = qw(
     @default_data_fields
 );
 
-# NOTE: the 'sgNumber' field was purposely excluded
-# from the default field list for now.
 our @default_data_fields = qw (
     file
     a
@@ -68,6 +66,7 @@ our @default_data_fields = qw (
     nel
     sg
     sgHall
+    sgNumber
 
     commonname
     chemname
